#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SortingAlgorithms.h"

// Swap two values
void swap(Type *a, Type *b)
{
	Type temp = *b;
	*b = *a;
	*a = temp;
}


// Insertion Sort algorithm
void insertionSort(Type data[], int arraySize, BOOL (*compareFunc)(Type a, Type b))
{
	int i = 0,
		j = 0;
	Type currentValue = 0;

	for (i = 1; i < arraySize; ++i) {
		currentValue = data[i];
		j = i;
		// Shift the values before the current value one place to the left if the current value satisfies the comparison
		// function with them (e.g. less than, greater than etc)
		while ((j > 0) && (compareFunc(currentValue, data[j-1]) == TRUE)) {
			data[j] = data[j-1];
			--j;
		}
		data[j] = currentValue;
	}
}

// Selection Sort Algorithm
void selectionSort(Type data[], int arraySize, BOOL (*compareFunc)(Type a, Type b))
{
	int i = 0, j = 0;
	Type extremeValue = 0;

	// Iterate N times, finding the minimum (or maximum) value (depending on the comparison function)
	// and put it in its place
	for (i = 0; i < (arraySize - 1); i++) {
		extremeValue = i;
		for (j = i + 1; j < arraySize; j++) {
			if (compareFunc(data[j], data[extremeValue]) == TRUE)
				extremeValue = j;
		}
		swap(&data[i], &data[extremeValue]);
	}
}

// Quick Sorting Algorithm
void quickSort(Type data[], int arraySize, BOOL (*compareFunc)(Type a, Type b))
{
	q_sort(data, 0, arraySize - 1, compareFunc);
}

// Recursive function for the quick sort algorithm
void q_sort(Type data[], int left, int right, BOOL (*compareFunc)(Type a, Type b))
{
	Type pivotValue = (Type)0;		// The pivot point value
	int pivot;						// The pivot point index
	int l_hold, r_hold;				// Store the left and right indices before moving

	l_hold = left;
	r_hold = right;
	pivot = left;
	pivotValue = data[pivot];
	while (left < right) {
		// Find the right values that satisfy the comparison function and move them to the left side of the pivot point
		while ((compareFunc(pivotValue, data[right]) == TRUE) && (left < right))
			right--;
		if (left != right) {
			swap(&data[left], &data[right]);
			left++;
		}
		// Find the left values that satisfy the comparison function and move them to the right side point
		while ((compareFunc(pivotValue, data[left]) == FALSE) && (left < right))
			left++;
		if (left != right)
		{
			swap(&data[left], &data[right]);
			right--;
		}
	}
	// Set the leftmost data to hold the pivot value
	data[left] = pivotValue;
	pivot = left;
	left = l_hold;
	right = r_hold;
	// Sort the left part of the pivot
	if (left < pivot)
		q_sort(data, left, pivot - 1, compareFunc);
	// Sort the right part of the pivot
	if (right > pivot)
		q_sort(data, pivot + 1, right, compareFunc);
}


// Heap Sort alogirthm
void heapSort(Type data[], int arraySize, BOOL (*compareFunc)(Type a, Type b))
{
	int i = 0;

	// Build the Heap
	for( i = arraySize / 2; i >= 0; i-- )
		percolateDown(data, i, arraySize, compareFunc );

	// Delete First Element from the heap and put it in the end reducing the size of the heap by one each time
	for( i = arraySize - 1; i > 0; i-- ) {
		swap(&data[0], &data[i]);					// Move min-max element to the end of the array
		percolateDown(data, 0, i, compareFunc );	// Heapify the remaining data
	}
}

void percolateDown(Type data[], int i, int arraySize, BOOL (*compareFunc)(Type a, Type b))
{
	int child;
	Type temp;

	for( temp = data[i]; LEFT_CHILD(i) < arraySize; i = child ) {
		child = LEFT_CHILD(i);
		// Find the min child (or max depending on the comparison function)
		if((child != (arraySize - 1)) && 
		   (compareFunc(data[child + 1], data[child]) == FALSE ))
				child = child + 1;

		if(compareFunc(temp, data[child]) == TRUE )
			data[i] = data[child];
		else
			break;
	}
	data[i] = temp;
}

// Merge Sort Algorithm
void mergeSort(Type data[], int arraySize, BOOL (*compareFunc)(Type a, Type b))
{
	Type *tempArray = NULL;

	tempArray = (Type *)malloc(arraySize * sizeof( Type ));
	if (tempArray == NULL) {
		fprintf(stderr, "Not Enough Memory!!!\n");
		exit(1);
	}

	// Start the sorting
	m_sort( data, tempArray, 0, arraySize - 1, compareFunc);
	free(tempArray);
}

// Sort a subarray using merge sort
void m_sort(Type data[], Type tempArray[], int left, int right, BOOL (*compareFunc)(Type a, Type b))
{
	int center = 0;

	if( left < right )
	{
		center = (left + right) / 2;
		// Sort the left part
		m_sort( data, tempArray, left, center, compareFunc);
		// Sort the right part
		m_sort( data, tempArray, center + 1, right, compareFunc);
		// Merge the left and right parts
		merge( data, tempArray, left, center + 1, right, compareFunc);
	}
}

// Merge two sorted subarrays
void merge(Type data[], Type tempArray[ ], int leftPos, int rightPos, int rightEnd, BOOL (*compareFunc)(Type a, Type b))
{
	int i, leftEnd, numOfElements, tempPos;

	leftEnd = rightPos - 1;
	tempPos = leftPos;
	numOfElements = rightEnd - leftPos + 1;

	// Loop over the two sorted subarrays and merge them
	while((leftPos <= leftEnd) && (rightPos <= rightEnd))
		if (compareFunc(data[leftPos], data[rightPos]) == TRUE) {
			tempArray[tempPos++] = data[leftPos++];
		}
		else
			tempArray[ tempPos++ ] = data[ rightPos++ ];

	// Copy rest of first half
	while( leftPos <= leftEnd )
		tempArray[tempPos++] = data[leftPos++];
	// Copy rest of second half
	while( rightPos <= rightEnd ) 
		tempArray[tempPos++] = data[rightPos++];

	// Copy the temporary array back in the original array
	for( i = 0; i < numOfElements; i++, rightEnd-- )
		data[rightEnd] = tempArray[rightEnd];
}
